//
// Copyright (C) 2002 Henning Spruth
// Copyright 1997-2000 by Easy Software Products.

// Written by Henning Spruth <henning.spruth@epost.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>

#include <FL/Fl.H>
#include <FL/Fl_Group.H>
#include <FL/Fl_Scrollbar.H>
#include <FL/fl_draw.H>

#include "LinkArray.h"
#include "FontStack.h"
#include "LineDef.h"
#include "Para.h"
#include "ParaStack.h"
#include "Table.h"
#include "TableStack.h"
#include "HtmlData.h"
#include "VisLinkArray.h"
#include "HtmlDisplay.h"



#define SCROLLBAR_WIDTH 8


void HtmlDisplay::draw()
{
    Table *curTable;
    //printf("In HtmlDisplay::draw() x=%d y=%d w=%d, h=%d\n", x(), y(), w(), h());

    if(htmlData==NULL)
	return;

    draw_child(xscroll);
    draw_child(yscroll);

    if(!fl_not_clipped(x(),y(),w()-scrollbar_size, h()-scrollbar_size))
	return;


    draw_box(FL_DOWN_BOX, x(), y(), 
	     w() - scrollbar_size , 
	     h() - scrollbar_size,
	     FL_WHITE);
    fl_push_clip(x() + frame_size, 
		 y() + frame_size, 
		 w() - SCROLLBAR_WIDTH -2*frame_size , 
		 h() - SCROLLBAR_WIDTH -2*frame_size);

    viewx=xscroll.value();
    viewy=yscroll.value();

    //   viewx-=x();
    //viewy-=y();
    int thisx=x();
    int thisy=y();

    fl_color(FL_BLACK);

    // Clear all link boxes
    visibleLinks->clear();
    
    int i;
    // The maximum y coordinate which still is going to be drawn; this is
    // required to draw tables, where the y coordinates of elements inside
    // the table may not be monotonically increasing
    int viewMaxY=viewh+viewy;
    //printf("Drawing %d elements, viewy=%d viewh=%d\n", htmlData->getElemCount(), viewy, viewh);
    for(i=0; i<htmlData->getElemCount(); i++) {
	Elem *elem=htmlData->getElem(i);
//	if(frame_size+elem->y-viewy>viewh)

	// If I'm **certain** that everything below this is below the
	// bottom then I'm done
	if(elem->y/*+elem->h*/>viewMaxY) {
	    //printf("End draw at element %d\n", i);
	    break;
	}
	// don't draw elements that are over the top
	if(elem->y-viewy+elem->h<frame_size)
	    continue;
	// don't draw elements that are completely below the bottom
	if(elem->y+htmlData->getElemMaxSize(elem)>viewy+viewh)
	    continue;
	int x=frame_size+elem->x-viewx+thisx;
	int y=frame_size+elem->y-viewy+thisy;
	// Store biggest y coordinate we chose to display
	if(elem->y+elem->h>viewMaxY) {
	    viewMaxY=elem->y+elem->h;
	    //printf("viewMaxY is now %d\n", viewMaxY);
	}
	int maxSize = htmlData->getElemMaxSize(elem);
	switch(elem->type) {
	case Ttext:
	case Tpre:
	case Timg:
	    fl_font(elem->font, elem->size);
	    //printf("Drawing '%s' at %d,%d\n", elem->args, x,y+maxSize);
	    fl_draw(elem->args, x, y+maxSize);
	    // If this text has a link, draw an underline
	    if(elem->linkIndex>=0) {
		fl_line(x, y+maxSize+2, x+elem->w, y+maxSize+2);
		visibleLinks->add(x, y, elem->w, maxSize, 
			   htmlData->getLink(elem->linkIndex));
	    }
	    break;
	case Ttable:
	    //printf("Table, w=%d h=%d\n", elem->w, elem->h);
	    curTable = (Table *) elem->args;
	    if(curTable->borderWidth>0) {
		// Draw table border:
		int u=x+curTable->outerFrame;
		int v=y+curTable->outerFrame;
		int lw=curTable->w-2*curTable->outerFrame-curTable->borderWidth;
		int lh=curTable->h-2*curTable->outerFrame-curTable->borderWidth-1;
		// Vertical lines
		for (int j=0; j<curTable->colCount; j++) {
		    fl_line(u, v, u, v+lh);
		    u+=2*curTable->innerFrame+curTable->getColWidth(j)
			+curTable->borderWidth;
		}
		int uend=u;
		// horizontal lines
		u=x+curTable->outerFrame;
		for (int j=0; j<curTable->rowCount; j++) {
		    fl_line(u, v, u+lw, v);
		    v+=2*curTable->innerFrame+curTable->getRowHeight(j)
			+curTable->borderWidth;
		}
		int vend=v;
		u=x+curTable->outerFrame;
		v=y+curTable->outerFrame;
		fl_line(u, vend, uend, vend);
		fl_line(uend, v, uend, vend);
	    }
	    break;
	case Thr:
	    fl_line_style(0);
	    fl_line(frame_size, y, frame_size+vieww, y);
	    break;
	case Tli:
	    if(elem->listno==0) {
		fl_font(FL_SYMBOL, elem->size);
		//printf("Drawing bullet at %d,%d\n", x-3*elem->spcWidth,y+maxSize);
		fl_draw("\267", x-3*elem->spcWidth, y+maxSize);
	    } else {
		char buf[5];
		fl_font(elem->font, elem->size);
		sprintf(buf, "%d", elem->listno);
		//printf("Drawing number '%s' at %d,%d\n", buf, x-3*elem->spcWidth,y+maxSize);
		fl_draw(buf, x-3*elem->spcWidth, y+maxSize);
	    }
	    break;
	}
    }
    //printf("Drew %d elements\n", i);
    fl_pop_clip();
}
 


void HtmlDisplay::clear() {
    visibleLinks->clear();
}



//
// 'HtmlDisplay::handle()' - Handle events in the widget.
//

int				// O - 1 if we handled it, 0 otherwise
HtmlDisplay::handle(int event)	// I - Event to handle
{
  int		i;		// Looping var
  int		xx, yy;		// Adjusted mouse position
  Link	*link;		// Current link

  //printf("In handle, event=%d\n");
  switch (event)
  {
    case FL_MOVE :
    case FL_PUSH :
        xx = Fl::event_x() - x();
	yy = Fl::event_y() - y();
        if (xx<w()-scrollbar_size && yy<h()-scrollbar_size)
	  break;
  case FL_FOCUS:
        xx = Fl::event_x() - x();
	yy = Fl::event_y() - y();
        if (xx<w()-scrollbar_size && yy<h()-scrollbar_size)
	    return 1;
	return Fl_Group::handle(event);
	break;
    case FL_KEYBOARD:
	switch(Fl::event_key()) {
	case FL_Left:
	    if(leftHandler!=NULL)
		leftHandler();
	    return 1;
	    break;
	}
	return Fl_Group::handle(event);
	break;
    default :
	// Use the Fl_Group handler...
	return (Fl_Group::handle(event));
  }

  

  //printf("Handle: checking for links\n");
  link = visibleLinks->underCursor(Fl::event_x(), Fl::event_y());
  if (link!=NULL) {
      if (event == FL_MOVE)
	  fl_cursor(FL_CURSOR_HAND);
      else {
	  fl_cursor(FL_CURSOR_DEFAULT);
	  if(linkHandler!=NULL)
	      linkHandler(link->target);
	  else
	      printf("No link handler; would link to %s\n", link->target);
      }
  } else {
      fl_cursor(FL_CURSOR_DEFAULT);
  }
  return (1);
}


void HtmlDisplay::resize(int newx, int newy, int neww, int newh) {

    Fl_Widget::resize(newx, newy, neww, newh);
    // resize scroll bars
    xscroll.resize(newx, newy+newh-SCROLLBAR_WIDTH, 
		   neww-SCROLLBAR_WIDTH, SCROLLBAR_WIDTH);
    yscroll.resize(newx+neww-SCROLLBAR_WIDTH, newy, 
		   SCROLLBAR_WIDTH, newh-SCROLLBAR_WIDTH);

    // update viewable portion
    viewh=newh-scrollbar_size-2*frame_size;
    vieww=neww-scrollbar_size-2*frame_size;

    // reformat window content
    htmlData->format(w()-2*frame_size-scrollbar_size);
    updateScrollbars();
    redraw();
}

void HtmlDisplay::load(const char *fileName) {

    //printf("Loading %s\n", fileName);
    htmlData->clear();
    clear();
    htmlData->loadFile(fileName);
    htmlData->parse();
    //htmlData->dump();
    htmlData->measure();
    htmlData->format(w()-2*frame_size-scrollbar_size);
    updateScrollbars();
    redraw();
}


//
// Callback functions for scrollbar
//

static void yscroll_callback(Fl_Widget *widget, void *)
{
    Fl_Scrollbar *scroll = (Fl_Scrollbar *) widget;
    int pos = scroll->value();
    ((HtmlDisplay *)(widget->parent()))->redraw();
    // ((HtmlDisplay *)(s->parent()))->topline(int(((Fl_Scrollbar*)s)->value()));
    //scroll.value(topline_, h(), 0, size_);
}

static void xscroll_callback(Fl_Widget *widget, void *)
{
    Fl_Scrollbar *scroll = (Fl_Scrollbar *) widget;
    int pos = scroll->value();
    ((HtmlDisplay *)(widget->parent()))->redraw();
    // ((HtmlDisplay *)(s->parent()))->topline(int(((Fl_Scrollbar*)s)->value()));
}


void HtmlDisplay::updateScrollbars() {
    xscroll.value(0, vieww, 0, htmlData->getPageW());
    yscroll.value(0, viewh, 0, htmlData->getPageH());
    //printf("updateScrollbars: vieww=%d pagew=%d viewh=%d, pageh=%d\n", vieww, htmlData->getPageW(), viewh, htmlData->getPageH());
}


//
// The Constructor
//
HtmlDisplay::HtmlDisplay(int x, int y, int w, int h, const char *label)
    : Fl_Group(x,y,w,h,label),
      yscroll(x+w-SCROLLBAR_WIDTH, y, SCROLLBAR_WIDTH, h-SCROLLBAR_WIDTH),
      xscroll(x, y+h-SCROLLBAR_WIDTH, w-SCROLLBAR_WIDTH, SCROLLBAR_WIDTH)
{

    yscroll.value(0, h-scrollbar_size, 0, 1);
    xscroll.type(FL_HORIZONTAL);
    xscroll.value(0, w-scrollbar_size, 0, 1);


    // default colors
    //color(FL_WHITE);
    //textcolor(FL_BLACK);
    //selection_color(FL_BLUE);

    viewh=h-scrollbar_size-2*frame_size;
    vieww=w-scrollbar_size-2*frame_size;

    visibleLinks = new VisLinkArray();

//  scrollbar_.step(8.0);
    yscroll.show();
    xscroll.show();
    yscroll.callback(yscroll_callback);
    xscroll.callback(xscroll_callback);

    htmlData = new HtmlData();

    linkHandler=NULL;
    leftHandler=NULL;

    end();
}



