//
// Copyright (C) 2002 Henning Spruth
// Copyright 1997-2000 by Easy Software Products.

// Written by Henning Spruth <henning.spruth@epost.de>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <sys/stat.h>
#include <sys/wait.h>

#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <string.h>
#include <FL/fl_ask.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Pack.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Menu_Bar.H>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <FL/Fl_Group.H>
#include <FL/Fl_Input.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Menu_Button.H>
#include <FL/Fl_Browser.H>
#include <FL/Fl_Return_Button.H>

#include "LinkArray.h"
#include "FontStack.h"
#include "LineDef.h"
#include "Para.h"
#include "ParaStack.h"
#include "Table.h"
#include "TableStack.h"
#include "HtmlData.h"
#include "VisLinkArray.h"
#include "HtmlDisplay.h"
#include "Url.h"

#include "History.h"
#include "http.h"

static char *versionString="tweb version 0.1";

Fl_Window* window=NULL;
Fl_Output* urlW=NULL;
HtmlDisplay *htmlW=NULL;
History *history=NULL;
Url *currentUrl=NULL;

static char bufUrlW[HTTP_MAX_URI];

static char *confDirName() {
    static char buf[1024];
    strcpy(buf, getenv("HOME"));
    strcat(buf, "/.tweb/");
    return buf;
}

static char *bookmarkFileName() {
    char *result=confDirName();
    strcat(result, "bookmarks.html");
    return result;
}



const char *handleZip() {
    static char tempfile[1024];
    char buf[256];

    strcpy(bufUrlW, currentUrl->getUrl());
    urlW->value(currentUrl->getUrl());
    urlW->redraw();

    sprintf(tempfile, "%s/%d.html",
	    getenv("TMPDIR") == NULL ? "/tmp" : getenv("TMPDIR"), getuid());
    const char *path=currentUrl->getPath();

    // strip leading slash from the file name
    while(*path=='/')
	path++;
    sprintf(buf, "unzip -p %s%s %s > %s",
	    confDirName(), currentUrl->getHost(), 
	    *path=='\0' ? "index.html" : path,
	    tempfile);

    pid_t pid = fork();
    if(pid==0) {
	//printf("buf='%s'\n",buf);
	execl("/bin/sh", "/bin/sh", "-c", buf, NULL);
    } else {
	int status;
	waitpid(pid, &status, 0);
	if(status!=0)
	    fl_alert("Exit status %d from\n%s", status, buf);
	return tempfile;
    }
}


const char *			// O - New filename or NULL
resolveUrl(const char *url)	// I - URL for access...
{

  char		method[HTTP_MAX_URI],	// URL method
		username[HTTP_MAX_URI],	// Username from URL
		hostname[HTTP_MAX_URI],	// Hostname from URL
		host[HTTP_MAX_URI],	// Host: field
		resource[HTTP_MAX_URI];	// Resource from URL
  int		port;			// Port number for method
  http_t	*http;			// HTTP connection
  http_status_t	status;			// HTTP status
  FILE		*fp;			// Temporary file
  char		buffer[8192];		// Copy buffer
  int		bytes;			// Number of bytes read
  int		tbytes;			// Total number of bytes read
  int		length;			// Content length
  static char	tempfile[1024];		// Temporary filename
  static char	auth[255] = "";		// Authentication string

  currentUrl->set(url);

  if (strncmp(url, "file:", 5) == 0 || strchr(url, ':') == NULL)
    return (url);


  if(strncmp(url, "zip:", 4)==0)
      return handleZip();

  httpSeparate(url, method, username, hostname, &port, resource);
  strcpy(host, hostname);
  //printf("Method: %s, hostname: %s, resource: %s\n", method, hostname, resource);

//  if (proxy_[0])
//  {
//    httpSeparate(proxy_, method, username, hostname, &port, resource);
//    strcpy(resource, url);
//  }

  if (strcmp(method, "http") != 0 && strcmp(method, "ipp") != 0)
    return (NULL);

  //set_status("Contacting host %s...", hostname);
  //printf("Contacting host %s...\n", hostname);
  sprintf(bufUrlW, "Accessing %s", hostname);
  urlW->value(bufUrlW);
  urlW->redraw();
  Fl::flush();

  if ((http = httpConnect(hostname, port)) == NULL)
  {
//    set_status(NULL);
    fl_alert("Unable to connect to %s:\n\n%s", hostname, strerror(errno));
    return (NULL);
  }

  if (username[0])
    httpEncode64(auth, username);

//  set_status("Host contacted, sending request...");
  //puts("Host contacted, sending request...");

  // Do HTTP GET requests until we succeed or error out...
  do
  {
    httpClearFields(http);
    httpSetField(http, HTTP_FIELD_HOST, host);

    if (auth[0])
    {
      sprintf(buffer, "Basic %s", auth);
      httpSetField(http, HTTP_FIELD_AUTHORIZATION, buffer);
    }

    if (httpGet(http, resource))
    {
      status = HTTP_UNAUTHORIZED;
      continue;
    }

    while ((status = httpUpdate(http)) == HTTP_CONTINUE);

    if (status == HTTP_UNAUTHORIZED)
    {
      const char *userpass;


      sprintf(buffer, "Authentication required.\nEnter username:password for %s:",
              host);
      if ((userpass = fl_input(buffer, username)) == NULL)
	break;

      httpEncode64(auth, userpass);
      httpFlush(http);
    }
  }
  while (status == HTTP_UNAUTHORIZED);

  //puts ("Writing to buffer");

  // Copy the output from the server to a single temp file (no caching for now)
  sprintf(tempfile, "%s/%d.html",
          getenv("TMPDIR") == NULL ? "/tmp" : getenv("TMPDIR"), getuid());
  if ((fp = fopen(tempfile, "wb")) == NULL)
  {
//    set_status(NULL);
    httpClose(http);
    return (NULL);
  }

  length = atoi(httpGetField(http, HTTP_FIELD_CONTENT_LENGTH));
  tbytes = 0;
  while ((bytes = httpRead(http, buffer, sizeof(buffer))) > 0)
  {
    tbytes += bytes;

//    if (length == 0)
//      set_status("Received %d bytes...", tbytes);
//    else
//      set_status("Received %d of %d bytes (%d%%)...", tbytes, length,
//        	 100 * tbytes / length);

    if (length == 0)
	sprintf(bufUrlW, "Got %d", tbytes);
    else
	sprintf(bufUrlW, "Got %d/%d", tbytes, length);
    urlW->value(bufUrlW);
    urlW->redraw();
    Fl::flush();

    fwrite(buffer, 1, bytes, fp);
  }

//  set_status(NULL);
  fclose(fp);
  //puts ("Closing HTTP connection");
  httpClose(http);
  strcpy(bufUrlW, url);
  urlW->value(bufUrlW);

  return (tempfile);
}

static void backCb(Fl_Widget *widget, void *data) {
    char *url;
    // remove current page from stack
    url=history->popUrl();
    free(url);
    // remove previous page from stack
    url=history->popUrl();
    if(url!=NULL) {
	history->pushUrl(url);
	htmlW->load(resolveUrl(url));
    }
}

static void handleLeft() {
    // This is called from the HtmlDisplay widget when the
    // left key is pressed.
    backCb(NULL, NULL);
}

static void addBookmarkCb(Fl_Widget *widget, void *data) {
    char buf[1024];
    struct stat statbuf;
    char *fn = confDirName();
    if(stat(fn, &statbuf)!=0) {
	if(errno==ENOENT) {
	    if(mkdir(fn, 0777)!=0) {
		sprintf(buf, "Can't create '%s': %s",
			fn, strerror(errno));
		fl_alert(buf);
		return;
	    }
	} else {
	    sprintf(buf, "Can't open '%s': %s",
		    fn, strerror(errno));
	    fl_alert(buf);
	    return;
	}
    }
    fn=bookmarkFileName();
    FILE *fp=fopen(fn, "a");
    if (fp==NULL) {
	sprintf(buf, "Can't open '%s': %s",
		fn, strerror(errno));
	fl_alert(buf);
	return;
    }
    const char *head=htmlW->getHead();
    if(head!=NULL) {
	// full format if there is an HTML head
	fprintf(fp, "<a href=\"%s\">%s</a>: %s<p>\n",
		currentUrl->getUrl(),
		head,
		currentUrl->getUrl());
    } else {
	// shorter format if no head is found
	fprintf(fp, "<a href=\"%s\">%s</a><p>\n",
		currentUrl->getUrl(),
		currentUrl->getUrl());

    }
    fclose(fp);
}

static void goBookmarkCb(Fl_Widget *widget, void *data) {
    htmlW->load(resolveUrl(bookmarkFileName()));
//    const char *newUrl = fl_input("Please enter the URL:", NULL);
//    htmlW->load(resolveUrl(newUrl));
}

static void openCb(Fl_Widget *widget, void *data) {
    const char *newUrl = fl_input("Please enter the URL:", NULL);
    htmlW->load(resolveUrl(newUrl));
}

static void exitCb(Fl_Widget *widget, void *data) {
    exit(0);
}

static void aboutCb(Fl_Widget *widget, void *data) {
    fl_message("tweb - a tiny web browser\nVersion "TWEB_VERSION"\n(c) 2002, Henning Spruth\nhttp://agtoys.sourceforge.net/tweb");
}




static Fl_Menu_Item menubar[] = {
//    {"@menu", 0,  NULL, 0,  FL_SUBMENU | FL_SYMBOL_LABEL, 0, 8, 0 },
    {"F", 0,  NULL, 0,   FL_SUBMENU, 0, 0, 8, 0 },
    {"&Open URL", 0,  openCb, 0, 0, 0, 0, 8, 0 },
    {"&Add Bookmark", 0,  addBookmarkCb, 0, 0, 0, 0, 8, 0 },
    {"&Bookmarks", 0,  goBookmarkCb, 0, 0, 0, 0, 8, 0 },
    {"A&bout", 0,  aboutCb, 0, 0, 0, 0, 8, 0 },
    {"E&xit", 0,  exitCb, 0, 0, 0, 0, 8, 0 },
    {0},
//    {"@<-", 0,  backCb, 0, 0, FL_SYMBOL_LABEL, 0, 8, 0 },
    {"B", 0,  backCb, 0, 0, 0, 0, 8, 0 },
    {0}
};



static void handleLink(const char *target) {
    char newUrl[HTTP_MAX_URI];
    
    //printf("Handling link to %s\n", target);
    if(strncmp(target, "http://", 7)==0
       || strncmp(target, "file:", 5)==0
       || strncmp(target, "zip:", 4)==0) {
	// Explicit URL with host, no problem
	strcpy(newUrl, target);
    } else if (strncmp(target, "//", 2)==0) {
	sprintf(newUrl, "http:%s", target);
    } else {
	switch(currentUrl->getType()) {
	case URL_HTTP:
	    sprintf(newUrl, "http://%s", currentUrl->getHost());
	    break;
	case URL_ZIP:
	    sprintf(newUrl, "zip:%s", currentUrl->getHost());
	    break;
	default:
	    newUrl[0]='\0';
	    break;
	}
	if(target[0]=='/')
	    strcat(newUrl, target);
	else if (target[0]=='#') {
	    // Reference to different tag on same page
	    strcat(newUrl, currentUrl->getPath());
	    char *s;
	    s=&newUrl[strlen(newUrl)];
	    // Search backwards in the existing URL and check if it
	    // already contained a #<target> marker
	    while(s!=newUrl && *s!='/' && *s!='#') s--;
	    if(*s=='#')
		*s='\0';	// yes, cut it off
	    // concatenate new target marker to old path
	    strcat(newUrl, target);
	} else {
	    // Count the number of updir references, if any
	    int upCount=0;
	    const char *s=target;
	    while(strncmp(s, "../", 3)==0) {
		s+=3;
		upCount++;
	    }
	    // seems to be a relative file name, so maintain the
	    // existing directory part
	    strcat(newUrl, currentUrl->getDir());
	    // If there are updir references, adjust the path 
	    // accordingly
	    while(upCount>0) {
		// locate the last slash of the directory part
		char *t=newUrl+strlen(newUrl);
		while(t>newUrl&&*t!='/') t--;
		// remove it
		*t='\0';
		upCount--;
	    }
	    strcat(newUrl, "/");
	    strcat(newUrl, s);
	}
    }

    //printf("Full URL is '%s'\n", newUrl);

    urlW->value(newUrl);
    history->pushUrl(newUrl);
    urlW->redraw();
    htmlW->load(resolveUrl(newUrl));
}

int main(int argc, char **argv) {

    int winWidth;
    int winHeight;

    history = new History();
#ifdef HOST
    winWidth=400;
    winHeight=600;
#else
    winWidth=160;
    winHeight=186;
#endif

    window= new Fl_Window(winWidth, winHeight, "tweb");

    Fl_Menu_Bar* menuW =  new Fl_Menu_Bar(0, 0, 40, 14);
    menuW->menu(menubar);
    menuW->labelsize(8);

    urlW = new Fl_Output(40,0,winWidth-40,14);
    urlW->textsize(8);
    urlW->value("FOO");
    urlW->align(FL_ALIGN_RIGHT);


    htmlW = new HtmlDisplay(0, 14, winWidth, winHeight-14, "");
//    htmlW->link(handleLink);
//  htmlW->show();
    htmlW->setLinkHandler(handleLink);
    htmlW->setLeftHandler(handleLeft);

    currentUrl = new Url();

    if(argc>=2) {
	//puts("Before load");
	//char *url=argv[1];
	char *url=argv[1];
	history->pushUrl(url);
	htmlW->load(resolveUrl(url));
	//puts("Before 2nd show");
    }
    
//    Html *html = new Html();
    //html->loadFile("/home/henning/agenda/mystuff/tweb/try/www.sueddeutsche.de/handheld/index.php3?id=263");
//    html->loadFile(argv[1]);
//    html->parse();
    //html->dump();
//    html->measure();
//    html->format(winWidth-2*htmlW->frame_size-htmlW->scrollbar_size);
    
//    htmlW->setContent(html);


    window->resizable(htmlW);
    window->show();
  //puts("Before run");



  Fl::run();

  delete window;
}
